"""Comparison engine for spark connectors"""

from pyspark.sql import DataFrame, Window
from pyspark.sql.functions import col, lower, trim, abs, when, lit, row_number
from pyspark.sql.types import NumericType
from engines.compare_engine import CompareEngine

class CompareEngineSpark(CompareEngine):
    """Comparison engine for spark connectors"""

    def __init__(self, df_source: DataFrame, df_expected: DataFrame, options: dict):
        """Initialize the CompareEngineSpark class"""

        self.df_source = df_source
        self.df_expected = df_expected
        self.options = options
        self.mode = options["compare_mode"].upper()

    def compare(self) -> bool:
        """Compare the source and expected datasets"""
        if self.mode == "ORDER":
            if self.df_source.count() == 0 and self.df_expected.count() == 0 and self.options["allow_no_rows"]:
                return True
            if not self.compare_structure():
                return False
            if not self.compare_length():
                return False
            if not self.compare_data_with_order():
                return False
            return True

    def compare_length(self) -> bool:
        """Compare the length of the source and expected datasets"""
        count_source = self.df_source.count()
        count_expected = self.df_expected.count()

        if count_source != count_expected:
            self.error_type = "count"
            self.error_message = f"The count in source dataset ({count_source}) is different than the count in the expected dataset ({count_expected})"
            return False

        return True

    def compare_structure(self) -> bool:
        """Compare the structure of the source and expected datasets"""
        # Normalize column names to lowercase
        self.df_source = self.df_source.select(*[col(c).alias(c.lower()) for c in self.df_source.columns])
        self.df_expected = self.df_expected.select(*[col(c).alias(c.lower()) for c in self.df_expected.columns])

        # Remove columns specified in the ignore options
        if self.options is not None and self.options["ignore"] is not None:
            columns_to_ignore = [col.lower() for col in self.options["ignore"]]

            self.df_source = self.df_source.drop(*columns_to_ignore)
            self.df_expected = self.df_expected.drop(*columns_to_ignore)

        # Compare column headers
        source_cols = set(self.df_source.columns)
        expected_cols = set(self.df_expected.columns)

        missing_columns = expected_cols - source_cols
        if missing_columns:
            self.error_type = "headers"
            self.error_message = f"Missing columns: {', '.join(missing_columns)}"
            return False

        extra_columns = source_cols - expected_cols
        # remove extra columns from source
        if extra_columns:
            self.df_source = self.df_source.drop(*extra_columns)

        return True

    def compare_data_with_order(self) -> bool:
        """Compare the source and expected datasets with "order" mode"""

        # Sort the data based on the specified columns
        if self.options is not None and self.options["sort"] is not None:
            sort_columns = self.options["sort"]
            sort_columns = [x.lower() for x in sort_columns]

            if len(self.options["sort"]) > 0 and self.options["sort"][0] == "*":
                sort_columns = self.df_expected.columns
        else:
            sort_columns = self.df_expected.columns

        df_window = Window.orderBy(*sort_columns)
        self.df_source = self.df_source.withColumn("__ploosh_order_key", row_number().over(df_window))
        self.df_expected = self.df_expected.withColumn("__ploosh_order_key", row_number().over(df_window))

        tolerance = self.options["tolerance"]

        # Trim the data if the trim option is specified
        if self.options["trim"]:
            for col_name in self.df_source.columns:
                self.df_source = self.df_source.withColumn(col_name, trim(col(col_name)))
            for col_name in self.df_expected.columns:
                self.df_expected = self.df_expected.withColumn(col_name, trim(col(col_name)))

        # Convert all string values to lowercase if the case_insensitive option is specified
        if self.options["case_insensitive"]:
            for col_name in self.df_source.columns:
                self.df_source = self.df_source.withColumn(col_name, lower(col(col_name)))
            for col_name in self.df_expected.columns:
                self.df_expected = self.df_expected.withColumn(col_name, lower(col(col_name)))

        # Add a suffix to the column names to avoid conflicts
        df1 = self.df_source.select([col(c).alias(f"{c}_source") for c in self.df_source.columns])
        df2 = self.df_expected.select([col(c).alias(f"{c}_expected") for c in self.df_source.columns])

        # Clean up the order key column
        df1 = df1.withColumnRenamed("__ploosh_order_key_source", "__ploosh_order_key")
        df2 = df2.withColumnRenamed("__ploosh_order_key_expected", "__ploosh_order_key")

        # Join the two dataframes on the order key
        joined_df = df1.join(df2, on=["__ploosh_order_key"], how="inner")
        joined_df = joined_df.drop("__ploosh_order_key")

        # Drop the order key column from the source and expected dataframes
        self.df_source = self.df_source.drop("__ploosh_order_key")
        self.df_expected = self.df_expected.drop("__ploosh_order_key")

        diff_columns = []
        for c in self.df_source.columns:
            source_col = f"{c}_source"
            expected_col = f"{c}_expected"
 
            source_is_numeric = isinstance(df1.schema[source_col].dataType, NumericType)
            expected_is_numeric = isinstance(df2.schema[expected_col].dataType, NumericType)

            # Check if the column is numeric and the tolerance is greater than 0
            if source_is_numeric and expected_is_numeric and tolerance > 0:
                filters = (abs(col(source_col) - col(expected_col)) > tolerance)
            else:
                filters = (col(source_col) != col(expected_col))

            # Add state columns to show the differences
            diff_columns.append(when(filters, lit(False)).alias(f"{c}_state"))

            # Add the source and expected columns to the diff_columns
            diff_columns.append(when(filters, col(source_col)).alias(source_col))
            diff_columns.append(when(filters, col(expected_col)).alias(expected_col))

        # Filter the joined dataframe to get the differences
        result = joined_df.select(*diff_columns).filter(
            " OR ".join([f"{c}_state = FALSE" for c in self.df_source.columns])
        ).drop(*[f"{c}_state" for c in self.df_source.columns])

        datasource_count = self.df_source.count()

        # Convert the result to a pandas dataframe for the output report
        df_differences = result.toPandas()

        # Set the success rate and error message if there are differences
        if len(df_differences) > 0:
            self.success_rate = (datasource_count - len(df_differences)) / datasource_count
            if self.success_rate < self.options["pass_rate"]:
                self.error_message = "Some rows are not equal between source dataset and expected dataset"
                self.error_type = "data"

            self.df_compare_gap = df_differences.reindex(sorted(df_differences.columns), axis=1)

            return False

        # Set the error message if the source and expected datasets are empty
        if self.error_message is None and datasource_count == 0 and not self.options["allow_no_rows"]:
            self.error_message = "Source and exptected datasets are empty but no allow_no_rows option is set to False"
            self.error_type = "data"

            return False

        return True



"""Automatized Testing Framework"""

import sys
from colorama import Fore
from pyspark.sql import SparkSession
from case import StateStatistics
from connectors import get_connectors
from exporters import get_exporters
from parameters import Parameters
from configuration import Configuration
from logs import Log, print_compare_state, print_summary


def load_data(current_case, process_type, statistics):
    """Load data from source or expected"""
    try:
        # Attempt to load data for the current case
        current_case.load_data(process_type)
        return True
    except Exception as e:
        # Handle any errors that occur during data loading
        current_case.load_data_error(process_type, str(e))
        current_case.calculate_durations()
        statistics.add_state(current_case.state)
        Log.print_error(str(e))
    return False


def compare_data(current_case, statistics, spark_session):
    """Compare data between source and expected"""
    try:
        # Compare data using Spark if both connectors are Spark-based
        if current_case.source.connector.is_spark and current_case.expected.connector.is_spark:
            current_case.compare_dataframes_with_spark(spark_session)
        else:
            # Otherwise, use a standard comparison
            current_case.compare_dataframes()
        statistics.add_state(current_case.state)
        return True
    except Exception as e:
        # Handle any errors that occur during data comparison
        current_case.compare_dataframes_error(str(e))
        current_case.calculate_durations()
        statistics.add_state(current_case.state)
        Log.print_error(str(e))
    return False


def execute(args=None, spark_session=None):
    """Main function to execute test cases"""
    Log.init()
    Log.print_logo()

    statistics = StateStatistics()

    Log.print(f"{Fore.CYAN}Initialization[...]")
    try:
        # Parse command-line arguments
        if args is None:
            parameters = Parameters(sys.argv)
        else:
            parameters = Parameters(args)

        # Initialize Spark session if needed
        if parameters.spark_mode is True and spark_session is None:
            Log.print("Start spark session")
            spark_session = SparkSession.builder \
                .master("local") \
                .appName("ploosh") \
                .getOrCreate()

        # Load connectors and exporters
        Log.print("Load connectors")
        connectors = get_connectors(spark_session)
        Log.print("Load exporters")
        exporters = get_exporters()

        # Load configuration and test cases
        Log.print("Load configuration")
        configuration = Configuration(parameters, connectors, exporters)
        cases = configuration.get_cases()
    except Exception as e:
        # Handle any errors that occur during initialization
        Log.print_error(str(e))
        sys.exit(1)

    Log.print(f"{Fore.CYAN}Start processing tests cases[...]")
    for i, case_name in enumerate(cases):
        current_case = cases[case_name]

        # Skip disabled test cases
        if current_case.disabled:
            Log.print(f"{Fore.MAGENTA}{case_name} [...] ({i + 1}/{len(cases)}) - Skipped")
            statistics.add_state(current_case.state)
            continue

        Log.print(f"{Fore.MAGENTA}{case_name} [...] ({i + 1}/{len(cases)}) - Started")

        # Load source data
        Log.print("Load source data")
        if not load_data(current_case, "source", statistics):
            continue

        # Load expected data
        Log.print("Load expected data")
        if not load_data(current_case, "expected", statistics):
            continue

        # Compare source and expected data
        Log.print("Compare source and expected data")
        if not compare_data(current_case, statistics, spark_session):
            continue

        # Print comparison state and calculate durations
        print_compare_state(current_case)
        current_case.calculate_durations()

    Log.print(f"{Fore.CYAN}Export results[...]")
    # Export test results
    configuration.exporter.export(cases)
    Log.print(f"{Fore.CYAN}Summary[...]")
    # Print summary of test results
    print_summary(cases, statistics)

    # Exit with error code if there were errors and failure_on_error is set
    if statistics.error > 0 and parameters.failure_on_error:
        sys.exit(1)



"""Initialization for command line"""

import os
import sys

# Add the current directory to the system path
sys.path.append(os.path.dirname(__file__))

from execute import execute


def execute_cases(
    cases=None,
    connections=None,
    spark=None,
    spark_session=None,
    filter=None,
    path_output=None,
):
    """Execute test cases with the given parameters"""
    args = ["ploosh"]

    # Add cases parameter to arguments if provided
    if cases is not None:
        args.append("--cases")
        args.append(cases)

    # Add connections parameter to arguments if provided
    if connections is not None:
        args.append("--connections")
        args.append(connections)

    # Add spark parameter to arguments if provided
    if spark is not None:
        args.append("--spark")
        args.append(spark)

    # Add filter parameter to arguments if provided
    if filter is not None:
        args.append("--filter")
        args.append(filter)

    # Add output path parameter to arguments if provided
    if path_output is not None:
        args.append("--output")
        args.append(path_output)

    # Execute the test cases with the constructed arguments
    execute(args, spark_session)



import numpy as np
import pandas as pd

from engines.load_engine import LoadEngine

class LoadEngineNative(LoadEngine):
    """Load engine for native Pandas"""
    def __init__(self, configuration, options, connection):
        """Initialize the LoadEngineNative class"""

        self.configuration = configuration
        self.options = options
        self.connection = connection

    def execute(self, df_data):
        """Execute the load engine"""

        self.count = len(df_data)

        # Cast columns to specified types
        for column in self.options["cast"]:
            column_name = self.get_insensitive_item(column["name"], df_data.columns)
            column_type = column["type"]
            if column_type == "datetime":
                column_type = "datetime64[ns]"
            df_data[column_name] = df_data[column_name].astype(column_type, errors="ignore")

        # Remap bad columns type
        for column in df_data.select_dtypes(include=["object"]).columns:
            if len(df_data) == 0:
                continue

            if type(df_data[column][0]).__name__ == "Decimal":
                df_data[column] = df_data[column].astype(float, errors="ignore")

        # Remove time zones
        date_columns = df_data.select_dtypes(include=["datetime64[ns, UTC]"]).columns
        for date_column in date_columns:
            df_data[date_column] = df_data[date_column].dt.tz_localize(None)
        self.count = len(df_data)

        return df_data



# pylint: disable=R0903,W0613
"""Base class for all load engines"""

class LoadEngine:
    """Base class for all load engines"""

    count = None
    configuration = None
    options = None
    connection = None
    df_data = None

    def get_insensitive_item(self, name: str, items: list) -> str:
        """Get item from list case-insensitively"""
        for item in items:
            if name.upper().strip() == item.upper().strip():
                return item
        return name

    def execute(self):
        """Execute the load engine"""
        return None



"""Comparison engine for native connectors"""

import numpy as np
import pandas as pd

from engines.compare_engine import CompareEngine

class CompareEngineNative(CompareEngine):
    """Comparison engine for native connectors"""

    def __init__(self, df_source, df_expected, options):
        """Initialize the CompareEngineNative class"""

        self.df_source = df_source
        self.df_expected = df_expected
        self.options = options
        self.mode = options["compare_mode"].upper()

    def compare(self) -> bool:
        """Compare the source and expected datasets"""

        if self.mode == "ORDER":
            if len(self.df_source) == 0 and len(self.df_expected) == 0 and self.options["allow_no_rows"]:
                return True
            if not self.compare_structure():
                return False
            if not self.compare_length():
                return False
            if not self.compare_data_with_order():
                return False
            return True

    def compare_length(self) -> bool:
        """Compare the length of the source and expected datasets"""

        count_source = len(self.df_source)
        count_expected = len(self.df_expected)

        if count_source != count_expected:
            self.error_type = "count"
            self.error_message = f"The count in source dataset ({count_source}) is different than the count in the expected dataset ({count_expected})"
            return False

        return True

    def compare_structure(self) -> bool:
        """Compare the structure of the source and expected datasets"""

        # Normalize column names to lowercase
        self.df_source.columns = map(str.lower, self.df_source.columns)
        self.df_expected.columns = map(str.lower, self.df_expected.columns)

        # Remove columns specified in the ignore options
        if self.options is not None and self.options["ignore"] is not None:
            columns_to_ignore = [x.lower() for x in self.options["ignore"]]

            self.df_source = self.df_source.drop(columns=columns_to_ignore, axis=1, errors="ignore")
            self.df_expected = self.df_expected.drop(columns=columns_to_ignore, axis=1, errors="ignore")

        # Sort headers before comparison
        df_columns_source = pd.DataFrame({"columns": self.df_source.columns}).sort_values(by=["columns"]).reset_index(drop=True)
        df_columns_expected = pd.DataFrame({"columns": self.df_expected.columns}).sort_values(by=["columns"]).reset_index(drop=True)

        # Compare column headers
        mask = ~df_columns_expected[["columns"]].apply(tuple, axis=1).isin(df_columns_source[["columns"]].apply(tuple, axis=1))
        missing_columns = df_columns_expected[mask]

        # Set error message if missing columns are found in the source
        if not missing_columns.empty:
            missing_columns_list = ", ".join(missing_columns["columns"].to_list())
            self.error_type = "headers"
            self.error_message = f"Missing columns: {missing_columns_list}"
            return False

        # Remove columns that are not in the expected dataframe
        self.df_source = self.df_source[df_columns_expected["columns"]]
        self.df_expected = self.df_expected[df_columns_expected["columns"]]

        return True

    def compare_data_with_order(self) -> bool:
        """Compare the source and expected datasets with "order" mode"""

        # Sort the datasets if the sort option is specified
        if self.options is not None and self.options["sort"] is not None:
            sort_columns = self.options["sort"]
            sort_columns = [x.lower() for x in sort_columns]

            if len(self.options["sort"]) > 0 and self.options["sort"][0] == "*":
                sort_columns = self.df_expected.columns["columns"].to_list()

            self.df_source = self.df_source.sort_values(by=sort_columns).reset_index(drop=True)
            self.df_expected = self.df_expected.sort_values(by=sort_columns).reset_index(drop=True)

        tolerance = self.options["tolerance"]

        df_source_preprocessed = self.df_source.copy()
        df_expected_preprocessed = self.df_expected.copy()

        datasource_count = len(self.df_source)

        # Trim the data if the trim option is specified
        if self.options["trim"]:
            df_source_preprocessed = df_source_preprocessed.applymap(lambda x: x.strip() if isinstance(x, str) else x)
            df_expected_preprocessed = df_expected_preprocessed.applymap(lambda x: x.strip() if isinstance(x, str) else x)

        # Convert all string values to lowercase if the case_insensitive option is specified
        if self.options["case_insensitive"]:
            df_source_preprocessed = df_source_preprocessed.applymap(lambda x: x.lower() if isinstance(x, str) else x)
            df_expected_preprocessed = df_expected_preprocessed.applymap(lambda x: x.lower() if isinstance(x, str) else x)

        def compare_with_tolerance(val1, val2):
            """Compare two values with tolerance"""
            if str(val1.dtype) in ("float64", "Int64") and str(val2.dtype) in ("float64", "Int64"):
                return abs(val1 - val2) > tolerance
            return val1 != val2

        # Compare the source and expected datasets with tolerance
        diff_with_tol = df_source_preprocessed.combine(df_expected_preprocessed, compare_with_tolerance)

        # Get the indices of the differences
        diff_indices = np.where(diff_with_tol)

        # Create a dataframe to store the differences
        df_differences = pd.DataFrame(index=self.df_source.index, columns=pd.MultiIndex.from_tuples([(col, "source") for col in self.df_source.columns] + [(col, "expected") for col in self.df_source.columns]))

        # Fill the differences dataframe with the values from the source and expected datasets
        for i, j in zip(*diff_indices):
            col = self.df_source.columns[j]
            df_differences.loc[self.df_source.index[i], (col, "source")] = self.df_source.iat[i, j]
            df_differences.loc[self.df_source.index[i], (col, "expected")] = self.df_expected.iat[i, j]

        # Drop rows and columns with all NaN values
        df_differences = df_differences.dropna(axis=1, how="all").dropna(axis=0, how="all")

        # Set the success rate and error message
        if len(df_differences) > 0:
            self.success_rate = (datasource_count - len(df_differences)) / datasource_count
            if self.success_rate < self.options["pass_rate"]:
                self.error_message = "Some rows are not equal between source dataset and expected dataset"
                self.error_type = "data"

            self.df_compare_gap = df_differences.reindex(sorted(df_differences.columns), axis=1)

            return False

        # Set the error message if the source and expected datasets are empty
        if self.error_message is None and datasource_count == 0 and not self.options["allow_no_rows"]:
            self.error_message = "Source and expected datasets are empty but no allow_no_rows option is set to False"
            self.error_type = "data"

            return False

        return True



"""Data connectors"""
from importlib import import_module
from logs import Log
import inspect
import os


def get_connectors(spark_session):
    """Get all existing connectors"""

    connectors = {}

    # List all Python files in the current directory that start with "connector_"
    files = [
        name
        for name in os.listdir(os.path.dirname(__file__))
        if name.endswith(".py") and name.startswith("connector_")
    ]

    for file in files:
        module_name = file[:-3]  # Remove the ".py" extension to get the module name

        try:
            # Import the module dynamically
            for name, obj in inspect.getmembers(import_module(f"connectors.{module_name}")):
                if inspect.isclass(obj) and name.startswith("Connector"):
                    current_connector = obj()  # Instantiate the connector class

                    # If a Spark session is provided and the connector is Spark-based, set the Spark session
                    if spark_session is not None and current_connector.is_spark:
                        current_connector.spark = spark_session

                    # Add the connector to the connectors dictionary
                    connectors[current_connector.name] = current_connector
        except Exception as e:
            Log.print_warning(f"Could not load connector {module_name}")
            Log.print_warning(str(e))


    return connectors



# pylint: disable=R0903
"""Connector to read SQL file"""

from connectors.connector import Connector


class ConnectorSQLSpark(Connector):
    """Connector to execute SQL query over Spark"""

    def __init__(self):
        # Initialize the connector with its name and indicate it uses Spark
        self.name = "SQL_SPARK"
        self.is_spark = True
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "query"}  # SQL query to execute
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Execute the SQL query using Spark and return the resulting DataFrame
        df = self.spark.sql(configuration["query"])

        return df



# pylint: disable=R0903
"""Connector to read Databricks database"""

import pandas as pd
from sqlalchemy import create_engine
from connectors.connector import Connector


class ConnectorDatabricks(Connector):
    """Connector to read Databricks database"""

    def __init__(self):
        # Initialize the connector with its name and connection definitions
        self.name = "DATABRICKS"
        self.connection_definition = [
            {
                "name": "token",  # Token for authentication
            },
            {
                "name": "hostname",  # Hostname of the Databricks instance
            },
            {
                "name": "database",  # Database name
            },
            {
                "name": "http_path",  # HTTP path for the Databricks cluster
            },
            {
                "name": "port",  # Port number (default is 443)
                "default": 443,
                "type": "integer",
            },
        ]
        self.configuration_definition = [
            {"name": "query"},  # SQL query to execute
            {"name": "connection"},  # Connection name
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Extract connection parameters
        token = connection["token"]
        hostname = connection["hostname"]
        database = connection["database"]
        port = connection["port"]
        http_path = connection["http_path"]

        # Create the connection string for Databricks
        connection_string = (
            f"databricks://token:{token}@{hostname}:{port}/{database}?http_path={http_path}"
        )

        # Create a SQLAlchemy engine using the connection string
        sql_connection = create_engine(connection_string, echo=False)

        # Execute the SQL query and read the data into a pandas DataFrame
        df = pd.read_sql(configuration["query"], sql_connection)

        return df



# pylint: disable=R0903
"""Connector to read ODBC connection"""

import warnings
import pandas as pd
import pyodbc
from connectors.connector import Connector


class ConnectorODCB(Connector):
    """Connector to read ODBC connection"""

    def __init__(self):
        # Initialize the connector with its name and connection definitions
        self.name = "ODBC"
        self.connection_definition = [
            {
                "name": "mode",
                "default": "DSN",
                "validset": ["DSN", "connection_string"],
            },
            {
                "name": "DSN",  # Data Source Name for the ODBC connection
                "default": None
            },
            {
                "name": "connection_string",
                "default": None
            },
            {
                "name": "auto_commit",
                "type": "boolean",
                "default": True,  # Whether to enable auto-commit
            },
            {
                "name": "use_credentials",
                "type": "boolean",
                "default": False,  # Whether to use credentials for the connection
            },
            {
                "name": "user",
                "default": None,  # Username for the connection
            },
            {
                "name": "password",
                "default": None,  # Password for the connection
            },
            {
                "name": "encoding",
                "default": "UTF-8",  # Encoding to use for the connection
            },
        ]
        self.configuration_definition = [{"name": "query"}, {"name": "connection"}]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        if connection["mode"] == "DSN":
            # Establish the ODBC connection using the provided DSN and optional credentials
            if connection["use_credentials"]:
                odbc_connection = pyodbc.connect(
                    f"DSN={connection['DSN']}",
                    user=connection["user"],
                    password=connection["password"],
                    autocommit=connection["auto_commit"],
                )
            else:
                odbc_connection = pyodbc.connect(
                    f"DSN={connection['DSN']};", autocommit=connection["auto_commit"]
                )
        else:
            odbc_connection = pyodbc.connect(
                connection["connection_string"], autocommit=connection["auto_commit"]
            )

        # Suppress warnings related to encoding settings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)

            # Set the encoding for the ODBC connection
            odbc_connection.setdecoding(
                pyodbc.SQL_CHAR, encoding=connection["encoding"]
            )
            odbc_connection.setdecoding(
                pyodbc.SQL_WCHAR, encoding=connection["encoding"]
            )
            odbc_connection.setencoding(encoding=connection["encoding"])

            # Execute the SQL query and read the data into a pandas DataFrame
            df = pd.read_sql(configuration["query"], odbc_connection)

        return df



# pylint: disable=R0903
"""Connector to read Semantic Model from Fabric XMLA endpoint"""

import pandas as pd
import requests
from azure.identity import ClientSecretCredential, InteractiveBrowserCredential, UsernamePasswordCredential
from connectors.connector import Connector
import json

class ConnectorSemanticModel(Connector):
    """Connector to read Semantic Model using Fabric XMLA endpoint"""

    def __init__(self):
        self.name = "SEMANTIC_MODEL"
        self.connection_definition = [
            {
                "name": "mode",
                "default": "oauth",
                "validset": ["oauth"]  # , "token", "spn"] To add once tested
            },
            {
                "name": "token",
                "default": None
            },
            {
                "name": "tenant_id",
                "default": None
            },
            {
                "name": "client_id",
                "default": None
            },
            {
                "name": "client_secret",
                "default": None
            },
            {
                "name": "dataset_id"
            }
        ]
        self.configuration_definition = [
            {
                "name": "query"
            },
            {
                "name": "body",
                "default": None
            }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        mode = connection["mode"]
        dataset_id = connection["dataset_id"]
        query = configuration["query"]

        if mode == "oauth":
            try:
                interactive_browser_credential_class = InteractiveBrowserCredential()
                scope = 'https://analysis.windows.net/powerbi/api/.default'
                access_token_class = interactive_browser_credential_class.get_token(scope)
                token_string = access_token_class.token
            except Exception as connection_error:
                raise ValueError(connection_error)

        # uses the token provided in the connection_definition
        elif mode == "token":
            token_string = connection["token"]

        # get a token from a registered azure app
        elif mode == "spn":
            scope = 'https://analysis.windows.net/powerbi/api/.default'
            tenant_id = connection["tenant_id"]
            client_id = connection["client_id"]
            client_secret = connection["client_secret"]
            authority = f'https://login.microsoftonline.com/'
            credential = ClientSecretCredential(tenant_id, client_id, client_secret, authority=authority)
            token = credential.get_token(scope)
            token_string = token.token  # need to define header

        # Initialize query
        post_query = f'https://api.powerbi.com/v1.0/myorg/datasets/{dataset_id}/executeQueries'
        header = {'Authorization': f'Bearer {token_string}', 'Content-Type': 'application/json'}
        body = '''{
        "queries": [
            {
            "query": "%s"
        }
        ],
        "serializerSettings": {
            "includeNulls": "true"
        }
        }''' % (query)

        post_r = requests.post(url=post_query, data=body, headers=header)

        if post_r.status_code == 200:
            output = post_r.json()
            df_results = pd.DataFrame(output)
            df_tables = pd.DataFrame(df_results["results"][0])
            df_rows = pd.DataFrame(df_tables["tables"][0])
            flatten_data = df_rows.values.flatten()
            df = pd.json_normalize(flatten_data)  # type: ignore

            return df

        elif post_r.status_code == 400:
            response = json.loads(post_r.text)
            error_code = response['error']['code']
            error_message = response['error']['pbi.error']['details'][0]['detail']['value']
            raise ValueError(f"DAX Execution Error : {error_code}\n{error_message}")

        elif post_r.status_code == 404:
            raise ValueError("Connection issue: PowerBIEntityNotFound")

        else:
            raise ValueError("Execution Error")



"""Module to manage the configuration of the framework"""
import copy
import pathlib
import yaml
from pyjeb import control_and_setup
from pyjeb.exception import InvalidParameterException
from case import Case, ConnectionDescription
from parameters import Parameters


class Configuration:
    """Manage configuration"""
    parameters = None
    connectors = None
    exporter = None
    connections = {}

    # Definition of the structure of a test case
    case_definition = [
        {"name": "disabled", "type": "boolean", "default": False},
        {"name": "options", "type": "dict", "default": {}},
        {"name": "options.compare_mode", "type": "string", "validset": ["order"], "default": "order"},
        {"name": "options.ignore", "type": "list", "default": None},
        {"name": "options.sort", "type": "list", "default": None},
        {"name": "options.cast", "type": "list", "default": []},
        {"name": "options.cast.name", "type": "string"},
        {
            "name": "options.cast.type",
            "type": "string",
            "validset": ["float", "int", "datetime", "string"],
        },
        {"name": "options.pass_rate", "type": "decimal", "default": 1},
        {"name": "options.allow_no_rows", "type": "boolean", "default": True},
        {"name": "options.trim", "type": "boolean", "default": False},
        {"name": "options.case_insensitive", "type": "boolean", "default": False},
        {"name": "options.tolerance", "type": "decimal", "default": 0},
        {"name": "source", "type": "dict"},
        {"name": "source.type"},
        {"name": "source.connection", "default": None},
        {"name": "expected", "type": "dict"},
        {"name": "expected.connection", "default": None},
        {"name": "expected.connection", "default": None},
    ]

    def __init__(self, parameters: Parameters, connectors: dict, exporters: dict):
        """Initialize Configuration with parameters, connectors, and exporters"""
        self.parameters = parameters
        self.connectors = connectors

        # Initialize connections and set the exporter
        self.init_connections()
        self.set_exporter(exporters)

    def get_module_type(self, obj: dict):
        """Get type of module from user configuration"""
        if "type" not in obj.keys():
            raise InvalidParameterException("Property 'type' is not defined")

        module_type = str(obj["type"]).upper()
        if module_type not in self.connectors.keys():
            raise InvalidParameterException(
                f"Property 'type' ({module_type}) is not a valid connector"
            )

        return module_type

    def get_connection(self, name):
        """Get connection of module from user configuration"""
        if name is None:
            return None

        if name not in self.connections.keys():
            raise InvalidParameterException(f"Connection {name} do not exists")

        return self.connections[name]

    def init_connections(self):
        """Load connections and apply control and variable setup"""
        if self.parameters.path_connection is None:
            self.connections = {}
        else:
            # Load connections from the specified path
            with open(self.parameters.path_connection, encoding="UTF-8") as file:
                content = file.read()
                self.connections = yaml.safe_load(content)

        # Copy connections and initialize each one
        connections = copy.copy(self.connections)
        for connection_name in connections:
            connection = connections[connection_name]
            module_type = self.get_module_type(connection)

            connection_definition = self.connectors[module_type].connection_definition

            # Apply control and setup to the connection
            self.connections[module_type] = control_and_setup(
                connection, connection_definition, self.parameters.variables, None
            )

    def get_cases(self):
        """Get cases from user configuration and apply control and variable setup"""
        cases = {}
        # Get list of case files based on the specified filter
        files_list = list(
            pathlib.Path(self.parameters.path_cases).rglob(
                self.parameters.path_cases_filter
            )
        )

        for file_path in files_list:
            with open(file_path, encoding="UTF-8") as file:
                configurations = yaml.load(file, Loader=yaml.loader.SafeLoader)
                for case_name in configurations.keys():
                    configuration = configurations[case_name]

                    # Apply control and setup to the case configuration
                    case = control_and_setup(
                        configuration,
                        self.case_definition,
                        self.parameters.variables,
                        None,
                    )

                    # Get and setup source configuration
                    source_module_type = self.get_module_type(configuration["source"])
                    source_connector = self.connectors[source_module_type]
                    source_configuration_definition = (
                        source_connector.configuration_definition
                    )

                    case["source"] = control_and_setup(
                        case["source"],
                        source_configuration_definition,
                        self.parameters.variables,
                        None,
                    )
                    source_definition = ConnectionDescription(
                        source_connector, self.get_connection(case["source"]["connection"])
                    )

                    # Get and setup expected configuration
                    expected_module_type = self.get_module_type(configuration["expected"])
                    expected_connector = self.connectors[expected_module_type]
                    expected_configuration_definition = (
                        expected_connector.configuration_definition
                    )

                    case["expected"] = control_and_setup(
                        case["expected"],
                        expected_configuration_definition,
                        self.parameters.variables,
                        None,
                    )
                    expected_definition = ConnectionDescription(
                        expected_connector,
                        self.get_connection(case["expected"]["connection"]),
                    )

                    # Ensure both source and expected connectors are either Spark or non-Spark
                    if source_connector.is_spark != expected_connector.is_spark:
                        raise InvalidParameterException(
                            f"'{case_name}': Source and expected must be both Spark or not Spark connectors"
                        )

                    # Add the case to the cases dictionary
                    cases[case_name] = Case(
                        configuration,
                        source_definition,
                        expected_definition,
                        case["options"],
                        case["disabled"],
                    )

        return cases

    def set_exporter(self, exporters):
        """Set exporter from args"""
        if self.parameters.export not in exporters.keys():
            raise InvalidParameterException(
                f"Parameter 'export' ({self.parameters.export}) is not a valid exporter"
            )

        # Set the exporter and its output path
        self.exporter = exporters[self.parameters.export]
        self.exporter.output_path = self.parameters.path_output



# pylint: disable=R0903
"""Connector to read BigQuery database"""

import pandas as pd
import pandas_gbq
from sqlalchemy import create_engine
from connectors.connector import Connector


class ConnectorBigQuery(Connector):
    """Connector to read BigQuery database"""

    def __init__(self):
        # Initialize the connector with its name and connection definitions
        self.name = "BIGQUERY"
        self.connection_definition = [
            {
                "name": "credentials",  # Credentials for authentication
                "default": None,
            },
            {
                "name": "credentials_type",  # Type of credentials (service account or current user)
                "validset": ["service_account", "current_user"],
                "default": "service_account",
            },
            {
                "name": "project_id",  # Project ID for BigQuery
                "default": None,
            },
        ]
        self.configuration_definition = [
            {"name": "query"},  # SQL query to execute
            {"name": "connection"},  # Connection name
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""
        # Extract credentials and credentials type from the connection
        credentials = connection["credentials"]
        credentials_type = connection["credentials_type"]

        # If using service account credentials, create a connection string and use SQLAlchemy
        if credentials_type == "service_account":
            connection_string = f"bigquery://?credentials_base64={credentials}"
            sql_connection = create_engine(connection_string, echo=False)
            df = pd.read_sql(configuration["query"], sql_connection)
        # If using current user credentials, use pandas_gbq to read the data
        elif credentials_type == "current_user":
            df = pandas_gbq.read_gbq(
                configuration["query"], connection["project_id"], progress_bar_type=None
            )

        return df



# pylint: disable=R0903
"""Connector to read JSON file"""

import pandas as pd
from connectors.connector import Connector


class ConnectorJSON(Connector):
    """Connector to read JSON file"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "JSON"
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"},  # Path to the JSON file
            {"name": "encoding", "type": "string", "default": "utf-8"},  # Encoding to use when reading the JSON file.
            {"name": "lines", "type": "boolean", "default": False},  # Whether to treat the file as line-delimited JSON (one JSON object per line).
            {"name": "nrows", "type": "integer", "default": None}  # Number of lines to read from a line-delimited JSON file.
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Read the JSON file using pandas with the specified delimiter
        df = pd.read_json(configuration["path"],
                         encoding = configuration["encoding"],
                         lines = configuration["lines"],
                         nrows = configuration["nrows"]
                         )
        return df



# pylint: disable=R0903
"""Connector to return empty"""

from pyspark.sql.types import StructType
from connectors.connector import Connector


class ConnectorEmptySpark(Connector):
    """Connector to return empty"""

    def __init__(self):
        # Initialize the connector with its name and indicate it uses Spark
        self.name = "EMPTY_SPARK"
        self.is_spark = True
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = []  # No specific configuration parameters required

    def get_data(self, configuration: dict, connection: dict):
        """Return empty value"""

        # Create an empty RDD (Resilient Distributed Dataset)
        empty_rdd = self.spark.sparkContext.emptyRDD()

        # Define an empty schema (no columns)
        columns = StructType([])

        # Create an empty DataFrame using the empty RDD and schema
        df = self.spark.createDataFrame(data = empty_rdd, schema = columns)

        return df



# pylint: disable=R0903
"""Connector to read Parquet file"""

from connectors.connector import Connector


class ConnectorParquetSpark(Connector):
    """Connector to read Parquet file"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "PARQUET_SPARK"
        self.is_spark = True
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"}  # Path to the Parquet file
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Extract the path and configuration parameters
        path = configuration["path"]

        # Read the Parquet file using pandas
        df = self.spark.read.parquet(path)

        return df



"""Automatized Testing Framework"""

from execute import execute


def main():
    """Entry point for conda execution"""
    # Call the main execution function
    execute()


# Check if the script is being run directly
if __name__ == "__main__":
    # Call the main execution function
    main()



# pylint: disable=R0903
"""Connector to read PostgreSQL database"""

import pandas as pd
from sqlalchemy import create_engine
import urllib
from connectors.connector import Connector


class ConnectorPostgreSQL(Connector):
    """Connector to read PostgreSQL database"""

    def __init__(self):
        # Initialize the connector with its name and connection definitions
        self.name = "POSTGRESQL"
        self.connection_definition = [
            {
                "name": "mode",
                "default": "password",
                "validset": ["password", "connection_string"],
            },
            {
                "name": "hostname",
                "default": None,
            },
            {
                "name": "database",
                "default": None,
            },
            {
                "name": "username",
                "default": None,
            },
            {
                "name": "password",
                "default": None,
            },
            {
                "name": "port",
                "default": 5432,
                "type": "integer",
            },
            {
                "name": "ssl_context",
                "default": False,
                "type": "boolean",
            },
            {
                "name": "connection_string",
                "default": None,
            },
        ]
        self.configuration_definition = [{"name": "query"}, {"name": "connection"}]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Use the provided connection string if mode is "connection_string"
        connection_string = connection["connection_string"]
        if connection["mode"] == "password":
            # Extract connection parameters
            port = connection["port"]
            hostname = connection["hostname"]
            username = connection["username"]
            password = connection["password"]
            database = connection["database"]
            # Create the connection string for PostgreSQL
            password = urllib.parse.quote_plus(password)
            connection_string = (
                f"postgresql+pg8000://{username}:{password}@{hostname}:{port}/{database}"
            )

        # Additional connection arguments
        connect_args = {}
        if connection["ssl_context"]:
            connect_args = {"ssl_context": True}

        # Create a SQLAlchemy engine using the connection string and additional arguments
        sql_connection = create_engine(
            connection_string, echo=False, connect_args=connect_args
        )

        # Execute the SQL query and read the data into a pandas DataFrame
        df = pd.read_sql(configuration["query"], sql_connection)

        return df



# pylint: disable=R0903
"""Connector to read Parquet file"""

import pandas as pd
from connectors.connector import Connector


class ConnectorParquet(Connector):
    """Connector to read Parquet file"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "PARQUET"
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"},  # Path to the Parquet file
            {"name": "columns", "type": "list", "default": None},  # Subset of columns to load
            {"name": "engine", "type": "string", "validset": ["auto", "pyarrow", "fastparquet"], "default": "auto"},  # Parquet engine to use ('auto', 'pyarrow', 'fastparquet')
            {"name": "filters", "type": "list", "default": None},  # Row group filters to apply (for 'pyarrow')
            {"name": "filters.column", "type": "string"},  # The name of the column to filter
            {"name": "filters.operator", "type": "string", "validset": ["==", "=", ">", ">=", "<", "<=", "!="]},  # The operator to be used
            {"name": "filters.value", "type": "integer"},  # The value to be used to filter the column
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Extract the path and configuration parameters
        path = configuration["path"]
        columns = configuration["columns"]
        engine = configuration["engine"]
        filters = configuration["filters"]
        list_filters = None
        if filters is not None:
            list_filters = (
                [(filter_spec["column"], filter_spec["operator"], filter_spec["value"]) for filter_spec in filters]
                if filters else None
            )

        # Read the Parquet file using pandas
        df = pd.read_parquet(path,
                             columns=columns,
                             engine=engine,
                             filters=list_filters)
        return df



"""Module to manage test case"""
from dataclasses import dataclass
from datetime import datetime
import numpy as np
from engines.compare_engine_native import CompareEngineNative
from engines.compare_engine_spark import CompareEngineSpark
from engines.load_engine_native import LoadEngineNative
from engines.load_engine_spark import LoadEngineSpark

@dataclass
class StateStatistics:
    """Statistics of test case executions"""
    not_executed = 0
    executed = 0
    passed = 0
    failed = 0
    error = 0
    total = 0

    def add_state(self, state):
        """Add new state to statistics"""
        if state == "passed":
            self.passed += 1
        if state == "failed":
            self.failed += 1
        if state == "error":
            self.error += 1
        if state == "notExecuted":
            self.not_executed += 1

        if state != "notExecuted":
            self.executed += 1

        self.total += 1


@dataclass
class ConnectionDescription:
    """Tuple of connection and connector"""
    connector = None
    connection = None

    def __init__(self, connector, connection):
        self.connector = connector
        self.connection = connection


@dataclass
class Duration:
    """Structure of duration"""
    start = None
    end = None
    duration = None

    def calculate_duration(self):
        """Calculate the duration between start and end date"""
        if self.end is not None:
            duration = self.end - self.start
            self.duration = duration.seconds + (duration.microseconds / 1000000)


@dataclass
class CaseItem:
    """Structure of case item (source or expected)"""
    connector = None
    connection = None
    configuration = None
    duration = None
    df_data = None
    count = 0

    def __init__(self, configuration, connector, connection):
        self.duration = Duration()
        self.connector = connector
        self.connection = connection
        self.configuration = configuration


class Case:
    """Test case item"""
    options = None
    source = None
    expected = None
    global_duration = None
    compare_duration = None
    state = "notExecuted"
    error_type = None
    error_message = None
    df_compare_gap = None
    disabled = None
    success_rate = 1

    def __init__(self, configuration, source, expected, options, disabled):
        self.source = CaseItem(configuration["source"], source.connector, source.connection)
        self.expected = CaseItem(configuration["expected"], expected.connector, expected.connection)
        self.options = options
        self.disabled = disabled
        self.global_duration = Duration()
        self.compare_duration = Duration()

    def get_insensitive_item(self, name: str, items: list) -> str:
        """Get item from list case-insensitively"""
        for item in items:
            if name.upper().strip() == item.upper().strip():
                return item
        return name

    def load_data(self, obj_type: str):
        """Load data from connector"""
        if obj_type == "source":
            obj = self.source
        else:
            obj = self.expected

        obj.duration.start = datetime.now()

        if not self.source.connector.is_spark:
            load_engine = LoadEngineNative(obj.configuration, self.options, obj.connection)
        else:
            load_engine = LoadEngineSpark(obj.configuration, self.options, obj.connection)

        # Load data from connector
        obj.df_data = obj.connector.get_data(obj.configuration, obj.connection)

        # Execute load engine
        obj.df_data = load_engine.execute(obj.df_data)
        obj.count = load_engine.count

        obj.duration.end = datetime.now()

    def load_data_error(self, obj_type: str, message: str):
        """Setup error message for data loading"""
        if obj_type == "source":
            obj = self.source
        else:
            obj = self.expected

        self.state = "error"
        self.error_type = "data"
        self.error_message = message
        obj.duration.end = datetime.now()

    def compare_dataframes(self):
        """Compare source and expected dataframe"""
        self.compare_duration.start = datetime.now()

        compare_engine = CompareEngineNative(self.source.df_data, self.expected.df_data, self.options)
        compare_state = compare_engine.compare()

        self.error_message = compare_engine.error_message
        self.error_type = compare_engine.error_type
        self.df_compare_gap = compare_engine.df_compare_gap
        self.success_rate = compare_engine.success_rate

        self.compare_duration.end = datetime.now()

        if compare_state:
            self.state = "passed"
        else:
            self.state = "failed"

    def compare_dataframes_with_spark(self, spark_session):
        """Compare source and expected dataframe using Spark"""
        self.compare_duration.start = datetime.now()

        compare_engine = CompareEngineSpark(self.source.df_data, self.expected.df_data, self.options)
        compare_state = compare_engine.compare()

        self.error_message = compare_engine.error_message
        self.error_type = compare_engine.error_type
        self.df_compare_gap = compare_engine.df_compare_gap
        self.success_rate = compare_engine.success_rate

        self.compare_duration.end = datetime.now()

        if compare_state:
            self.state = "passed"
        else:
            self.state = "failed"
            self.state = "failed"

    def compare_dataframes_error(self, message):
        """Setup error message for compare engine"""
        self.state = "Error"
        self.error_type = "compare"
        self.error_message = message
        self.compare_duration.end = datetime.now()

    def calculate_durations(self):
        """Calculate durations"""
        self.source.duration.calculate_duration()
        self.expected.duration.calculate_duration()
        self.compare_duration.calculate_duration()

        ends = []
        if self.source.duration.end is not None:
            ends.append(self.source.duration.end)
        if self.expected.duration.end is not None:
            ends.append(self.expected.duration.end)
        if self.compare_duration.end is not None:
            ends.append(self.compare_duration.end)

        if len(ends) == 0:
            self.global_duration.duration = 0
        else:
            self.global_duration.start = self.source.duration.start
            self.global_duration.end = np.max(np.array(ends))
        self.global_duration.calculate_duration()



from engines.load_engine import LoadEngine

class LoadEngineSpark(LoadEngine):
    """Load engine for Spark"""

    def __init__(self, configuration, options, connection):
        """Initialize the LoadEngineSpark class"""

        self.configuration = configuration
        self.options = options
        self.connection = connection

    def execute(self, df_data):
        """Execute the load engine"""
        self.count = df_data.count()

        # Cast columns to specified types
        for column in self.options["cast"]:
            column_name = self.get_insensitive_item(column["name"], df_data.columns)
            column_type = column["type"]
            if column_type == "datetime":
                column_type = "timestamp"
                
            df_data = df_data.withColumn(column_name, df_data[column_name].cast(column_type))

        return df_data



# pylint: disable=R0903
"""Connector to read Snowflake database"""

import pandas as pd
from sqlalchemy import create_engine
from connectors.connector import Connector


class ConnectorSnowflake(Connector):
    """Connector to read Snowflake database"""

    def __init__(self):
        # Initialize the connector with its name and connection definitions
        self.name = "SNOWFLAKE"
        self.connection_definition = [
            {
                "name": "account_identifier",  # Snowflake account identifier
            },
            {
                "name": "username",  # Username for authentication
            },
            {
                "name": "password",  # Password for authentication
            },
            {
                "name": "database",  # Database name (optional)
                "default": None,
            },
            {
                "name": "schema",  # Schema name (optional)
                "default": None,
            },
            {
                "name": "warehouse",  # Warehouse name (optional)
                "default": None,
            },
            {
                "name": "role",  # Role name (optional)
                "default": None,
            },
        ]
        self.configuration_definition = [{"name": "query"}, {"name": "connection"}]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Extract connection parameters
        account_identifier = connection["account_identifier"]
        username = connection["username"]
        password = connection["password"]

        # Create the base connection string for Snowflake
        connection_string = f"snowflake://{username}:{password}@{account_identifier}/"

        # Append database and schema to the connection string if provided
        if connection["database"] is not None:
            connection_string += f"{connection['database']}/"
        if connection["schema"] is not None:
            connection_string += f"{connection['schema']}"

        # Add query parameters to the connection string
        connection_string += "?1=1"
        if connection["warehouse"] is not None:
            connection_string += f"&warehouse={connection['warehouse']}"
        if connection["role"] is not None:
            connection_string += f"&role={connection['role']}"

        # Create a SQLAlchemy engine using the connection string
        sql_connection = create_engine(connection_string, echo=False)

        # Execute the SQL query and read the data into a pandas DataFrame
        df = pd.read_sql(configuration["query"], sql_connection)

        return df



"""Result exporter"""
from importlib import import_module
import os
import inspect


def get_exporters():
    """Get all existing exporters"""
    connectors = {}

    # List all Python files in the current directory that start with "exporter_"
    files = [
        name
        for name in os.listdir(os.path.dirname(__file__))
        if name.endswith(".py") and name.startswith("exporter_")
    ]

    for file in files:
        module_name = file[:-3]  # Remove the ".py" extension to get the module name

        # Import the module dynamically
        module = import_module(f"exporters.{module_name}")

        # Inspect the module to find classes that start with "Exporter"
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and name.startswith("Exporter"):
                current_connector = obj()  # Instantiate the exporter class
                connectors[
                    current_connector.name
                ] = current_connector  # Add the exporter to the connectors dictionary

    return connectors



# pylint: disable=R0903
"""Connector to read CSV file"""

from connectors.connector import Connector


class ConnectorCSVSpark(Connector):
    """Connector to read CSV file with Spark"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "CSV_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = [
            {"name": "path", "type": "string"},  # Path to the CSV file
            {"name": "delimiter", "type": "string", "default": ","},  # Delimiter used in the CSV file
            {"name": "header", "type": "boolean", "default": True},  # Whether the CSV file has a header row
            {"name": "inferSchema", "type": "boolean", "default": False},  # Infers the input schema automatically from data
            {"name": "multiline", "type": "boolean", "default": False},  # Parse one record, which may span multiple lines, per file
            {"name": "quote", "type": "string", "default": '"'},  # Character used to denote the start and end of a quoted item
            {"name": "encoding", "type": "string", "default": 'UTF-8'},  # Encoding to use for UTF when reading/writing
            {"name": "lineSep", "type": "string", "default": "\n"},  # Character used to denote a line break
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Read the CSV file using Spark with the specified configuration options
        df = self.spark.read.option("delimiter", configuration["delimiter"])    \
                            .option("header", configuration["header"])          \
                            .option("inferSchema", configuration["inferSchema"])\
                            .option("multiline", configuration["multiline"])    \
                            .option("quote", configuration["quote"])            \
                            .option("encoding", configuration["encoding"])      \
                            .option("lineSep", configuration["lineSep"])        \
                            .csv(configuration["path"])

        return df



# pylint: disable=R0903,W0613
""" Base class for the comparison engines """

class CompareEngine:
    """Base class for the comparison engines"""

    success_rate = 1
    error_type = None
    error_message = None
    df_compare_gap = None
    df_source = None
    df_expected = None
    options = None
    mode = None

    def compare(self) -> bool:
        """Compare the source and expected datasets"""
        return False



import pandas as pd
from connectors.connector import Connector
from azure.identity import ClientSecretCredential
from sys import path
from pathlib import Path

class ConnectorAnalysisServices(Connector):
    """Connector to read Analysis Services Model using ADOMD"""

    def __init__(self):
        ## ADOMD.dll ##
            # Using dll file put into src\ploosh\connectors\modules
            # The file need to be packaged to work here
            # need to use the absPath (!!! check if it works on linux !!!)
        root = Path(r"\Program Files\Microsoft.NET\ADOMD.NET\\")
        adomd_path = str(max((root).iterdir()))
        path.append(adomd_path)
        #absPath = os.path.dirname(__file__)
        #path.append(absPath + '\\modules')

            # NEED to pip install pythonnet to make pyadomd work !!
        global Pyadomd
        from pyadomd import Pyadomd
        ## ADOMD.dll -- END ##

        self.name = "ANALYSIS_SERVICES"
        self.connection_definition = [
            {
                "name": "mode",
                "default": "oauth",
                "validset": ["oauth", "pbix"]  # , "token", "credentials", "spn"]
            },
            {
                "name": "token",
                "default": None
            },
            {
                "name": "username",
                "default": None
            },
            {
                "name": "password",
                "default": None
            },
            {
                "name": "tenant_id",
                "default": None
            },
            {
                "name": "client_id",
                "default": None
            },
            {
                "name": "client_secret",
                "default": None
            },
            {
                "name": "scope",
                "default": 'https://analysis.windows.net/powerbi/api/.default'
            },
            {
                "name": "dataset_id"
            },
            {
                "name": "server"
            }
        ]
        self.configuration_definition = [
            {
                "name": "query"
            }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        mode = connection["mode"]

        server = connection["server"]
        dataset_id = connection["dataset_id"]  # For local .pbix --> Dataset ID: in DAX Studio, right click to model name and choose "copy Database ID"

        query = configuration["query"]  # DAX Query

        # will open a login page in browser (if local AS instance, will connect automatically)
        if mode == "oauth":
            connection_string = f'Provider=MSOLAP;Data Source={server};Catalog={dataset_id};'

        # will open a login page in browser (if local AS instance, will connect automatically)
        elif mode == "pbix":
            connection_string = f'Provider=MSOLAP;Data Source={server};Catalog={dataset_id};'

        # uses the token provided in the connection_definition
        elif mode == "token":
            token = connection["token"]
            connection_string = f'Provider=MSOLAP;Data Source={server};Catalog={dataset_id};User Id=;Password={token};Impersonation Level=Impersonate;'

        # get a token from a registered azure app
        elif mode == "spn":
            scope = connection["scope"]
            tenant_id = connection["tenant_id"]
            client_id = connection["client_id"]
            client_secret = connection["client_secret"]
            authority = f'https://login.microsoftonline.com/'
            credential = ClientSecretCredential(tenant_id, client_id, client_secret)  # , authority=authority)
            token = credential.get_token(scope)
            token_string = token.token
            connection_string = f'Provider=MSOLAP;Data Source={server};Catalog={dataset_id};User Id=;Password={token_string};Impersonation Level=Impersonate;'

        # uses username and password
        elif mode == "credentials":
            username = connection["username"]
            password = connection["password"]
            connection_string = f'Provider=MSOLAP;Data Source={server};Catalog={dataset_id};User Id={username};Password={password};'

        # Create and open connection to AS instance
        con = Pyadomd(connection_string)
        try:
            con.open()  # Open the connection
        except:
            raise ValueError("Can't connect to the AS Instance")

        # execute DAX query
        with con.cursor() as cur:
            try:
                cur.execute(query)
                result = cur.fetchone()
                column_name = [i.name for i in cur.description]
                df = pd.DataFrame(result, columns=column_name)

                # Proactively close connection to AS instance
                con.close()

                return df
            except Exception as query_error:
                error_message = str(query_error)
                # Keep only error message without Technical Details
                error_summary = error_message.split("Technical Details")[0].strip().split("\r\n   at")[0].strip()
                raise Exception(f"Erreur lors de l'exécution de la requête :\n{str(error_summary)}")



# pylint: disable=R0903
"""Connector to read Delta table"""

from connectors.connector import Connector


class ConnectorDeltaSpark(Connector):
    """Connector to read Delta table with Spark"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "DELTA_SPARK"
        self.is_spark = True  # Indicates that this connector uses Spark
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"},  # Path to the Delta table
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Read the Delta table using Spark with the specified path
        df = self.spark.read.format("delta").load(configuration["path"])

        return df



"""Connector to read Excel file"""

import pandas as pd
from connectors.connector import Connector


class ConnectorExcel(Connector):
    """Connector to read Excel file"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "EXCEL"
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"},  # Path to the Excel file
            {"name": "sheet_name"},  # Sheet name
            {"name": "skiprows", "type": "integer", "default": 0},  # Number of rows to skip
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""
        # Read the Excel file using pandas with the specified configuration options
        df = pd.read_excel(
            configuration["path"],
            sheet_name=configuration["sheet_name"],
            skiprows=configuration["skiprows"],
        )
        return df



# pylint: disable=R0903
"""Connector to read CSV file"""

import json
import pandas as pd
from connectors.connector import Connector


class ConnectorCSV(Connector):
    """Connector to read CSV file"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "CSV"
        self.connection_definition = []  # No specific connection parameters required
        self.configuration_definition = [
            {"name": "path"},  # Path to the CSV file
            {"name": "delimiter", "default": ","},  # Delimiter used in the CSV file
            {"name": "infer", "type": "boolean", "default": True},  # Infer the column names
            {"name": "names", "type": "list", "default": None},  # Sequence of column labels to apply
            {"name": "usecols", "type": "list", "default": None},  # Subset of columns to select
            {"name": "skiprows", "type": "string", "default": None},  # Line numbers to skip (0-indexed) or number of lines to skip (int) at the start of the file
            {"name": "skipfooter", "type": "integer", "default": 0},  # Number of lines at bottom of file to skip (Unsupported with engine='c')
            {"name": "nrows", "type": "integer", "default": None},  # Number of rows of file to read. Useful for reading pieces of large files.
            {"name": "lineterminator", "type": "string", "default": None},  # Character used to denote a line break.
            {"name": "quotechar", "type": "string", "default": '"'},  # Character used to denote the start and end of a quoted item.
            {"name": "encoding", "type": "string", "default": "utf-8"},  # Encoding to use for UTF when reading/writing.
            {"name": "engine", "type": "string", "default": None},  # Parser engine to use.
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Extract the path and delimiter from the configuration
        path = configuration["path"]
        delimiter = configuration["delimiter"]
        header = None if configuration["infer"] is False else "infer"
        names = configuration["names"]
        usecols = configuration["usecols"]
        skiprows = None
        skipfooter = configuration["skipfooter"]
        nrows = configuration["nrows"]
        lineterminator = configuration["lineterminator"]
        quotechar = configuration["quotechar"]
        encoding = configuration["encoding"]
        engine = configuration["engine"]

        if configuration["skiprows"] is not None:
            try:
                skiprows = json.loads(configuration["skiprows"])
            except json.JSONDecodeError:
                raise ValueError("The variable is neither a list nor an integer.")

        if skiprows is not None and not isinstance(skiprows, (list, int)):
            raise ValueError("The variable is neither a list nor an integer.")

        # Read the CSV file using pandas with the specified delimiter
        df = pd.read_csv(path,
                         delimiter=delimiter,
                         header=header,
                         names=names,
                         usecols=usecols,
                         skiprows=skiprows,
                         skipfooter=skipfooter,
                         nrows=nrows,
                         lineterminator=lineterminator,
                         quotechar=quotechar,
                         engine=engine,
                         encoding=encoding)
        return df


# pylint: disable=R0903
"""Connector to read json file"""

from connectors.connector import Connector


class ConnectorJSONSpark(Connector):
    """Connector to read json file with Spark"""

    def __init__(self):
        # Initialize the connector with its name and configuration definitions
        self.name = "JSON_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = [
            {"name": "path", "type": "string"},  # Path to the JSON file
            {"name": "multiline", "type": "boolean", "default": True},  # Handles multi-line JSON files
            {"name": "encoding", "type": "string", "default": "UTF-8"},  # Character encoding format used in the JSON file
            {"name": "lineSep", "type": "string", "default": "\n"}  # Character used to denote a line break
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        # Read the JSON file using Spark with the specified configuration options
        df = self.spark.read.option("multiline", configuration["multiline"])    \
                            .option("encoding", configuration["encoding"])      \
                            .option("lineSep", configuration["lineSep"])        \
                            .json(configuration["path"])

        return df



# pylint: disable=R0903,W0613
"""Connector to access to remote data"""


class Connector:
    """Connector to access to remote data"""
    name = None  # Name of the connector
    connection_definition = None  # Definition of the connection parameters
    configuration_definition = None  # Definition of the configuration parameters
    is_spark = False  # Flag to indicate if the connector uses Spark
    spark = None  # Spark session object

    def get_data(self, configuration: dict, connection: dict):
        """Get data from connector"""
        return None  # This method should be overridden by subclasses to fetch data
